"""Activity class for testing."""


class Activity:
    """Activity class for testing."""

    def __init__(self, name: str = "basic", duration: int = 4) -> None:
        """Activity class for testing."""
        self.name = name
        self.duration = duration
